@@ -10,8 +10,6 @@ tool-augmented models using self-contained task bundles.
 
 import warnings
 
-from eval_protocol.adapters.braintrust import reward_fn_to_scorer, scorer_to_reward_fn
-
 from .auth import get_fireworks_account_id, get_fireworks_api_key
 from .common_utils import load_jsonl
 from .config import RewardKitConfig, get_config, load_config
@@ -24,6 +22,7 @@ from .mcp_env import (
     rollout,
     test_mcp,
 )
+from .data_loader import DynamicDataLoader, InlineDataLoader
 
 # Try to import FireworksPolicy if available
 try:
@@ -34,23 +33,69 @@ except (ImportError, AttributeError):
     _FIREWORKS_AVAILABLE = False
 # Import submodules to make them available via eval_protocol.rewards, etc.
 from . import mcp, rewards
-from .models import EvaluateResult, Message, MetricResult
+from .models import EvaluateResult, Message, MetricResult, EvaluationRow, InputMetadata
 from .playback_policy import PlaybackPolicyBase
 from .resources import create_llm_resource
 from .reward_function import RewardFunction
 from .typed_interface import reward_function
+from .quickstart import aha_judge, multi_turn_assistant_to_ground_truth, assistant_to_ground_truth
+from .pytest import evaluation_test, SingleTurnRolloutProcessor, RemoteRolloutProcessor
+from .pytest.parameterize import DefaultParameterIdGenerator
+
+try:
+    from .adapters import OpenAIResponsesAdapter
+except ImportError:
+    OpenAIResponsesAdapter = None
+
+try:
+    from .adapters import LangfuseAdapter, create_langfuse_adapter
+except ImportError:
+    LangfuseAdapter = None
+
+try:
+    from .adapters import BraintrustAdapter, create_braintrust_adapter
+except ImportError:
+    BraintrustAdapter = None
+
+try:
+    from .adapters import LangSmithAdapter
+except ImportError:
+    LangSmithAdapter = None
+
+# Remote server types
+from .types.remote_rollout_processor import (
+    InitRequest,
+    RolloutMetadata,
+    StatusResponse,
+    create_langfuse_config_tags,
+)
 
 warnings.filterwarnings("default", category=DeprecationWarning, module="eval_protocol")
 
 __all__ = [
+    "RemoteRolloutProcessor",
+    "InputMetadata",
+    "EvaluationRow",
+    "DefaultParameterIdGenerator",
+    "DynamicDataLoader",
+    "InlineDataLoader",
+    "aha_judge",
+    "multi_turn_assistant_to_ground_truth",
+    "assistant_to_ground_truth",
+    "evaluation_test",
+    "SingleTurnRolloutProcessor",
+    "OpenAIResponsesAdapter",
+    "LangfuseAdapter",
+    "create_langfuse_adapter",
+    "BraintrustAdapter",
+    "create_braintrust_adapter",
+    "LangSmithAdapter",
     # Core interfaces
     "Message",
     "MetricResult",
     "EvaluateResult",
     "reward_function",
     "RewardFunction",
-    "scorer_to_reward_fn",
-    "reward_fn_to_scorer",
     # Authentication
     "get_fireworks_api_key",
     "get_fireworks_account_id",
@@ -75,6 +120,11 @@ __all__ = [
     # Submodules
     "rewards",
     "mcp",
+    # Remote server types
+    "InitRequest",
+    "RolloutMetadata",
+    "StatusResponse",
+    "create_langfuse_config_tags",
 ]
 
 from . import _version
