@@ -9,11 +9,14 @@ import asyncio
 import hashlib
 import json
 import logging
+import time
 from contextlib import AsyncExitStack
-from typing import Any, Dict, List, Optional, Tuple
+from typing import Any, Dict, List, Optional, Tuple, cast
 
+import httpx
 from mcp.client.session import ClientSession
 from mcp.client.streamable_http import streamablehttp_client
+from mcp.types import Implementation
 
 from ...types import MCPSession
 
@@ -23,6 +26,10 @@ logger = logging.getLogger(__name__)
 class MCPConnectionManager:
     """Manages MCP client connections and session lifecycle."""
 
+    def __init__(self):
+        self._tools_cache: Dict[str, List[Dict]] = {}
+        self._tools_cache_lock = asyncio.Lock()
+
     async def initialize_session(self, session: MCPSession) -> None:
         """
         Initialize a persistent MCP session.
@@ -46,19 +53,26 @@ class MCPConnectionManager:
 
         exit_stack = AsyncExitStack()
 
-        client_info = None
-        if session.seed is not None or (session.dataset_row and session.dataset_row.environment_context):
-            from mcp.types import Implementation
+        # Attach client metadata for the server to consume (session_id, seed, config, etc.).
+        # The server inspects a private `_extra` dict on client_info, so we populate it here.
+        client_info = Implementation(name="reward-kit", version="1.0.0")
+        extra_data: Dict[str, Any] = {"session_id": session.session_id}
+        if session.seed is not None:
+            extra_data["seed"] = session.seed
+        if session.dataset_row and session.dataset_row.environment_context:
+            extra_data["config"] = session.dataset_row.environment_context
+        if session.dataset_row and session.dataset_row.id:
+            extra_data["dataset_row_id"] = session.dataset_row.id
+        if session.model_id:
+            extra_data["model_id"] = session.model_id
 
-            client_info = Implementation(name="reward-kit", version="1.0.0", _extra={})
-            if session.seed is not None:
-                client_info._extra["seed"] = session.seed
-            if session.dataset_row and session.dataset_row.environment_context:
-                client_info._extra["config"] = session.dataset_row.environment_context
-            if session.dataset_row and session.dataset_row.id:
-                client_info._extra["dataset_row_id"] = session.dataset_row.id
-            if session.model_id:
-                client_info._extra["model_id"] = session.model_id
+        # Merge with any existing _extra dict instead of overwriting
+        existing_extra = getattr(client_info, "_extra", None)
+        merged_extra: Dict[str, Any] = {}
+        if isinstance(existing_extra, dict):
+            merged_extra.update(existing_extra)
+        merged_extra.update(extra_data)
+        setattr(client_info, "_extra", merged_extra)
 
         read_stream, write_stream, _ = await exit_stack.enter_async_context(
             streamablehttp_client(session.base_url, terminate_on_close=True)
@@ -73,35 +87,59 @@ class MCPConnectionManager:
         session._mcp_session = mcp_session
         session._exit_stack = exit_stack
 
-        # Update session ID to match server's calculation (for control plane sync)
-        if client_info and hasattr(client_info, "_extra"):
-            extra_data = client_info._extra
-            if extra_data and isinstance(extra_data, dict):
+        # PRE-WARM: Discover and cache tools immediately after session initialization
+        # This prevents concurrent list_tools() calls later
+        await self._prewarm_tools_cache(session)
 
-                seed_value = extra_data.get("seed")
-                config_value = extra_data.get("config", {})
-                dataset_row_id_value = extra_data.get("dataset_row_id")
-                model_id_value = extra_data.get("model_id")
+    async def _prewarm_tools_cache(self, session: MCPSession) -> None:
+        """
+        Pre-warm the tools cache for this session's base URL.
+        This prevents concurrent list_tools() calls during discover_tools().
+        """
+        cache_key = session.base_url
 
-                stable_data = {
-                    "seed": seed_value,
-                    "config": config_value,
-                    "dataset_row_id": dataset_row_id_value,
-                    "model_id": model_id_value,
-                    "name": client_info.name,
-                    "version": client_info.version,
-                }
+        async with self._tools_cache_lock:
+            # Only fetch tools if not already cached for this base_url
+            if cache_key not in self._tools_cache:
+                logger.debug(f"Pre-warming tools cache for {cache_key}")
+                mcp_session_local = session._mcp_session
+                if mcp_session_local is None:
+                    raise RuntimeError("Session not initialized during prewarm")
+                tools_response = await mcp_session_local.list_tools()
+                tools = tools_response.tools if hasattr(tools_response, "tools") else []
 
-                stable_str = json.dumps(stable_data, sort_keys=True)
-                server_session_id = hashlib.md5(stable_str.encode()).hexdigest()
+                tool_schemas = []
+                for tool in tools:
+                    tool_schema = {
+                        "name": tool.name,
+                        "description": tool.description,
+                        "input_schema": (tool.inputSchema if hasattr(tool, "inputSchema") else {}),
+                    }
+                    tool_schemas.append(tool_schema)
 
-                # Update the session ID to match what the server generated
-                session.session_id = server_session_id
-                logger.debug(f"Updated session ID to match server: {server_session_id}")
+                self._tools_cache[cache_key] = tool_schemas
+                logger.debug(f"✅ PRE-WARMED {len(tool_schemas)} tools for{cache_key}")
+
+    async def reset_session(self, session: MCPSession) -> None:
+        """
+        Clean session data in remote mcp server for the given session
+        """
+        base_url = session.base_url.rstrip("/").removesuffix("/mcp")
+        url = f"{base_url}/control/reset_session"
+
+        headers = {"mcp-session-id": session.session_id}
+        body = {"seed": session.seed}
+
+        timeout = httpx.Timeout(15.0)
+        async with httpx.AsyncClient(timeout=timeout) as client:
+            resp = await client.post(url, headers=headers, json=body)
+            resp.raise_for_status()
+            logger.debug(f"Session {session.session_id}: reset_session -> {resp.json()}")
 
     async def discover_tools(self, session: MCPSession) -> List[Dict]:
         """
         Discover available tools from an MCP session.
+        Now uses pre-warmed cache to avoid concurrent list_tools() calls.
 
         Args:
             session: The MCPSession to discover tools from
@@ -112,9 +150,19 @@ class MCPConnectionManager:
         if not session._mcp_session:
             raise RuntimeError("Session not initialized")
 
+        cache_key = session.base_url
+
+        # Check cache first (should be pre-warmed during initialization)
+        async with self._tools_cache_lock:
+            if cache_key in self._tools_cache:
+                cached_tools = self._tools_cache[cache_key]
+                logger.debug(f"Using cached tools for session {session.session_id} ({len(cached_tools)} tools)")
+                return cached_tools
+
+        # Fallback: if cache miss (shouldn't happen with pre-warming), fetch directly
+        logger.warning(f"Cache miss for {cache_key} - this shouldn't happen with pre-warming")
         mcp_session = session._mcp_session
 
-        # Get available tools from MCP server
         tools_response = await mcp_session.list_tools()
         tools = tools_response.tools if hasattr(tools_response, "tools") else []
 
@@ -129,8 +177,26 @@ class MCPConnectionManager:
             }
             tool_schemas.append(tool_schema)
 
+        # Cache the result for future use
+        async with self._tools_cache_lock:
+            self._tools_cache[cache_key] = tool_schemas
+
         return tool_schemas
 
+    def clear_tools_cache(self, base_url: Optional[str] = None):
+        """
+        Clear the tools cache for debugging or when server tools change.
+
+        Args:
+            base_url: If provided, clear cache only for this URL. If None, clear all.
+        """
+        if base_url:
+            self._tools_cache.pop(base_url, None)
+            logger.debug(f"Cleared tools cache for {base_url}")
+        else:
+            self._tools_cache.clear()
+            logger.debug("Cleared all tools cache")
+
     async def get_initial_state(self, session: MCPSession) -> Any:
         """
         Get initial state from session-aware control plane endpoint.
@@ -149,8 +215,6 @@ class MCPConnectionManager:
         initial_observation = None
 
         try:
-            import httpx
-
             # Extract base URL and session ID from the MCP session
             base_url = session.base_url.rstrip("/").removesuffix("/mcp")
             session_id = session.session_id
@@ -160,8 +224,9 @@ class MCPConnectionManager:
 
                 # Query initial state endpoint
                 try:
-                    # Use shorter timeout for playback mode
-                    timeout = 3.0 if hasattr(session, "_is_playback_mode") and session._is_playback_mode else 5.0
+                    # Use shorter timeout for playback mode, longer timeout for high-concurrency initialization
+                    # (50+ concurrent sessions need more time for initial state setup)
+                    timeout = 3.0 if bool(getattr(session, "_is_playback_mode", False)) else 15.0
                     async with httpx.AsyncClient(timeout=timeout) as client:
                         initial_state_response = await client.get(
                             f"{base_url}/control/initial_state",
@@ -211,7 +276,10 @@ class MCPConnectionManager:
         try:
             # List available resources - this is where initial state should come from
             logger.debug(f"Session {session.session_id}: Discovering MCP resources for initial state...")
-            resources_response = await mcp_session.list_resources()
+            mcp_session_local = session._mcp_session
+            if mcp_session_local is None:
+                raise RuntimeError("Session not initialized while listing resources")
+            resources_response = await mcp_session_local.list_resources()
             resources = resources_response.resources if hasattr(resources_response, "resources") else []
             logger.debug(f"Session {session.session_id}: Found {len(resources)} MCP resources")
             for resource in resources:
@@ -238,17 +306,21 @@ class MCPConnectionManager:
                     f"Session {session.session_id}: Reading initial state from resource: {initial_state_resource.uri}"
                 )
 
-                resource_content = await mcp_session.read_resource(initial_state_resource.uri)
+                mcp_session_for_read = session._mcp_session
+                if mcp_session_for_read is None:
+                    raise RuntimeError("Session not initialized while reading resource")
+                resource_content = await mcp_session_for_read.read_resource(initial_state_resource.uri)
 
                 # Handle the new ResourceContents format
-                if hasattr(resource_content, "text"):
+                text_value = getattr(resource_content, "text", None)
+                if text_value is not None:
                     try:
-                        initial_observation = json.loads(resource_content.text)
+                        initial_observation = json.loads(text_value)
                         logger.info(
                             f"Session {session.session_id}: ✅ Successfully parsed JSON initial state with grid_layout: {initial_observation.get('grid_layout', 'N/A')[:20]}..."
                         )
                     except json.JSONDecodeError:
-                        initial_observation = {"observation": resource_content.text}
+                        initial_observation = {"observation": text_value}
                 elif (
                     hasattr(resource_content, "contents")
                     and resource_content.contents
@@ -256,11 +328,12 @@ class MCPConnectionManager:
                 ):
                     # Fallback to old format for backward compatibility
                     content = resource_content.contents[0]
-                    if hasattr(content, "text"):
+                    content_text = getattr(content, "text", None)
+                    if content_text is not None:
                         try:
-                            initial_observation = json.loads(content.text)
+                            initial_observation = json.loads(content_text)
                         except json.JSONDecodeError:
-                            initial_observation = {"observation": content.text}
+                            initial_observation = {"observation": content_text}
                     else:
                         initial_observation = {"observation": str(resource_content)}
                 else:
@@ -281,7 +354,10 @@ class MCPConnectionManager:
                         f"Session {session.session_id}: About to call mcp_session.read_resource with fallback URI: {first_resource.uri}"
                     )
 
-                    resource_content = await mcp_session.read_resource(first_resource.uri)
+                    mcp_session_for_fallback_read = session._mcp_session
+                    if mcp_session_for_fallback_read is None:
+                        raise RuntimeError("Session not initialized while reading fallback resource")
+                    resource_content = await mcp_session_for_fallback_read.read_resource(first_resource.uri)
 
                     logger.debug(
                         f"Session {session.session_id}: fallback read_resource returned type: {type(resource_content)}"
@@ -294,11 +370,12 @@ class MCPConnectionManager:
                     )
 
                     # Handle the new ResourceContents format
-                    if hasattr(resource_content, "text"):
+                    text_value_2 = getattr(resource_content, "text", None)
+                    if text_value_2 is not None:
                         try:
-                            initial_observation = json.loads(resource_content.text)
+                            initial_observation = json.loads(text_value_2)
                         except json.JSONDecodeError:
-                            initial_observation = {"observation": resource_content.text}
+                            initial_observation = {"observation": text_value_2}
                     elif (
                         hasattr(resource_content, "contents")
                         and resource_content.contents
@@ -306,11 +383,12 @@ class MCPConnectionManager:
                     ):
                         # Fallback to old format for backward compatibility
                         content = resource_content.contents[0]
-                        if hasattr(content, "text"):
+                        content_text_2 = getattr(content, "text", None)
+                        if content_text_2 is not None:
                             try:
-                                initial_observation = json.loads(content.text)
+                                initial_observation = json.loads(content_text_2)
                             except json.JSONDecodeError:
-                                initial_observation = {"observation": content.text}
+                                initial_observation = {"observation": content_text_2}
                         else:
                             initial_observation = {"observation": str(content)}
                     else:
@@ -363,9 +441,10 @@ class MCPConnectionManager:
         # Extract data plane results (observation only)
         if tool_result.content and len(tool_result.content) > 0:
             content = tool_result.content[0]
-            if hasattr(content, "text"):
+            text_value = getattr(content, "text", None)
+            if isinstance(text_value, str):
                 # Fix: Handle empty or invalid JSON responses gracefully
-                if not content.text or content.text.strip() == "":
+                if text_value.strip() == "":
                     logger.warning(f"Session {session.session_id}: Empty tool response from {tool_name}")
                     observation = {
                         "observation": "empty_response",
@@ -373,14 +452,14 @@ class MCPConnectionManager:
                     }
                 else:
                     try:
-                        observation = json.loads(content.text)
+                        observation = json.loads(text_value)
                     except json.JSONDecodeError as e:
                         logger.warning(
-                            f"Session {session.session_id}: Invalid JSON from {tool_name}: {content.text}. Error: {e}"
+                            f"Session {session.session_id}: Invalid JSON from {tool_name}: {text_value}. Error: {e}"
                         )
                         # Create a structured response from the raw text
                         observation = {
-                            "observation": content.text,
+                            "observation": text_value,
                             "session_id": session.session_id,
                             "error": "invalid_json_response",
                         }
@@ -405,9 +484,6 @@ class MCPConnectionManager:
         control_plane_info = {}
 
         try:
-            # Query control plane endpoints following the new architecture
-            import httpx
-
             # Extract base URL and session ID from the MCP session
             base_url = session.base_url.rstrip("/").removesuffix("/mcp")
             # Use the session ID from the established MCP session
@@ -490,10 +566,10 @@ class MCPConnectionManager:
                 await session._exit_stack.aclose()
             except asyncio.CancelledError:
                 # Handle cancellation gracefully (especially important for Python 3.12)
-                logger.debug(f"Session {session.session_id} close was cancelled")
+                logger.error(f"Session {session.session_id} close was cancelled")
             except Exception as e:
                 # Hitting this error, probably because of use of threads: "Attempted to exit cancel scope in a different task than it was entered in"
-                logger.debug(f"Error closing session {session.session_id}: {e}")
+                logger.error(f"Error closing session {session.session_id}: {e}")
             finally:
                 session._exit_stack = None
                 session._mcp_session = None
